(*
 * Various dependency graphs, sorting and theorem dependencies
 * $Id: dependency.sml 1162 2009-04-21 09:41:10Z maxim $
 *)

signature DEPENDENCY =
sig

type tdeps

val trans_dep : ('a * Symtab.key list) Symtab.table -> ('a * tdeps) Symtab.table
val sort   : ('a * tdeps) Symtab.table -> (Symtab.key * 'b) list -> (Symtab.key * 'b) list 
val extend : ('a * tdeps) Symtab.table -> ('a * Symtab.key list) Symtab.table -> ('a * tdeps) Symtab.table

val thms_dep_graph : (string * Thm.thm) list -> (string * Symtab.key list) Symtab.table

datatype Elem = Class | Tycon | Const

datatype ('a, 'b) Data = ClassData of 'a 
                       | TyconData of Datatype_Record.InstStruct
                       | ConstData of 'b 

type ('a, 'b) ElemData

val class_data : theory -> string -> ('a, 'b) ElemData
val tycon_data : theory -> string -> ('a, 'b) ElemData
val const_data : theory -> string -> ('a, 'b) ElemData

type TDeps
datatype ('a, 'b) ElemDataT = ElemDataT of { elem : Elem,
                                             ids : string list,
                                             data : ('a, 'b) Data option,
                                             t_deps : TDeps }
val trancl : ('a, 'b) ElemData list -> ('a, 'b) ElemDataT list
val t_sort : ('a, 'b) ElemDataT list -> ('a, 'b) ElemDataT list
val dep_consts_of_classes : ('a, 'b) ElemData list -> string list  (* consts on which classes depend on *)
val intro_consts_of_elems : ('a, 'b) ElemData list -> string list  (* consts which elements introduce *)

end


structure Dependency : DEPENDENCY =
struct

datatype Elem = Class | Tycon | Const

datatype ('a, 'b) Data = ClassData of 'a 
                       | TyconData of Datatype_Record.InstStruct
                       | ConstData of 'b 

datatype Deps = Deps of { dep_classes  : string list,
                          dep_tycons : string list,
                          dep_consts : string list }


datatype ('a, 'b) ElemData = ElemData of { elem : Elem,
                                           ids : string list,
                                           intro_consts : string list,
                                           data : ('a, 'b) Data option,
                                           deps : Deps }


datatype TDeps = TDeps of { dep_classes  : string list,
                            dep_tycons : string list,
                            dep_consts : string list }

datatype ('a, 'b) ElemDataT = ElemDataT of { elem : Elem,
                                             ids : string list,
                                             data : ('a, 'b) Data option,
                                             t_deps : TDeps }


fun dep_consts_of_classes [] = []
  | dep_consts_of_classes ((x as (ElemData{elem, deps=Deps {dep_consts, ...}, ...}))::xs) =
    (case elem of
      Class => Aux.union_str(dep_consts, dep_consts_of_classes xs)
    | _ => dep_consts_of_classes xs)

fun intro_consts_of_elems [] = []
  | intro_consts_of_elems ((x as (ElemData{elem, intro_consts, ...}))::xs) =
    Aux.union_str(intro_consts, intro_consts_of_elems xs)

fun class_deps thy cls =
 (let val alg = Aux.sort_alg thy
      val sup = Sorts.minimize_sort alg (Sorts.super_classes alg cls)
  in
   case (Aux.get_axclass_info thy cls) of 
     NONE => Deps { dep_classes = sup,
                    dep_tycons = [],
                    dep_consts = [] }
   | SOME info => (let val axs = #axioms info
                       val pm = #params info
                       fun from_axs (cl, t, c) [] = (cl, t, c)
                         | from_axs (cl, t, c) (x::xs) =
                            let val term = Aux.unvarify (prop_of x)
                                val (cl', t', c') = Aux.deps_of_term thy term
                            in
                            from_axs (Aux.union_str (cl, cl'), Aux.union_str (t, t'), Aux.union_str (c, c')) xs 
                            end
                       val (classes, tycons, consts) = from_axs (sup, [], []) axs
                       fun from_pm (cl, t, c) [] = (cl, t, c)
                         | from_pm (cl, t, c) ((s, ty)::xs) = 
                            let val cl' = Aux.all_classes_of_type ty
                                val t' = Aux.all_tycons_of_type ty
                            in
                             from_pm (Aux.union_str (cl, cl'), Aux.union_str (t, t'), c) xs
                            end 
                       val (classes', tycons', consts') = from_pm (classes, tycons, consts) pm
                    in
                     Deps { dep_classes = Aux.diff_str(classes', [cls]),
                            dep_tycons = tycons',
                            dep_consts = consts'}
                    end)
  end) 


fun class_data thy cls =
 let val d = class_deps thy cls
 in
  ElemData { elem = Class,
             ids = [cls],
             intro_consts = [(cls^"_class")],
             data = NONE,
             deps = d }
 end


fun datatype_data cls thy tc =
  (case (Datatype_Record.try_datatype thy tc) of
    NONE => NONE
  | SOME (Datatype_Record.Result {ids, data, class_deps, tycon_deps, ...}) => 
               (let val d = Deps { dep_classes = Aux.union_str (cls, class_deps),
                                   dep_tycons = tycon_deps,
                                   dep_consts = []}
                in
                 SOME (ElemData { elem = Tycon,
                                  ids = ids,
                                  intro_consts = [],
                                  data = SOME (TyconData data),
                                  deps = d })
                end))

fun record_data cls thy tc =
  (case (Datatype_Record.try_record thy tc) of
    NONE => NONE 
  | SOME (Datatype_Record.Result {ids, data, class_deps, tycon_deps, ...}) => 
               (let val d = Deps { dep_classes = Aux.union_str (cls, class_deps),
                                   dep_tycons = tycon_deps,
                                   dep_consts = []}
                in
                 SOME (ElemData { elem = Tycon,
                                  ids = ids,
                                  intro_consts = [],  (* this should be changed *)
                                  data = SOME (TyconData data),
                                  deps = d })
                end))

fun simple_tycon_data cls tc =
(let val d = Deps { dep_classes = cls,
                    dep_tycons = [],
                    dep_consts = []}
 in
  ElemData { elem = Tycon,
             ids = [tc],
             intro_consts = [],
             data = NONE,
             deps = d }
 end)

fun tycon_data thy tc =
 let val cls = Aux.all_classes_of_tycon thy tc
 in
   (if !AWE.skip_datatypes
    then (if !AWE.skip_records 
          then simple_tycon_data cls tc
          else (case (record_data cls thy tc) of
                      SOME data => data
                    | NONE => simple_tycon_data cls tc))
    else (if !AWE.skip_records
          then (case (datatype_data cls thy tc) of
                 SOME data => data
               | NONE => simple_tycon_data cls tc)
          else  
          (case (datatype_data cls thy tc) of
            SOME data => data
          | NONE => (case (record_data cls thy tc) of
                      SOME data => data
                    | NONE => simple_tycon_data cls tc))))      
 end




fun const_data thy c =
(case (Aux.const_type thy c) of
   NONE => error "Unexpected error in const_data"
 | SOME ty => (let val ty' = Aux.unvarifyT ty
                   val d = Deps { dep_classes = Aux.all_classes_of_type ty',
                                  dep_tycons = Aux.all_tycons_of_type ty',
                                  dep_consts = [] }
               in
                ElemData { elem = Const,
                           ids = [c],
                           intro_consts = [],
                           data = NONE,
                           deps = d }
               end))



(* transitive closure *)



fun lookup_deps [] (el, id) = NONE
  | lookup_deps ((ElemData{elem, ids, data, deps, ...})::xs) (el, id) = 
 (if (el = elem) then 
   (if (id mem_string ids) then SOME deps else lookup_deps xs (el, id))
  else lookup_deps xs (el, id))

fun deps_union (Deps {dep_classes, dep_tycons, dep_consts},
                Deps {dep_classes=dep_classes', dep_tycons=dep_tycons', dep_consts=dep_consts'}) =
Deps {dep_classes = Aux.union_str(dep_classes, dep_classes'),
      dep_tycons = Aux.union_str(dep_tycons, dep_tycons'),
      dep_consts = Aux.union_str(dep_consts, dep_consts')}



val empty_deps =  Deps {dep_classes = [],
                        dep_tycons = [],
                        dep_consts = []}

fun lookup_deps_union ids l = Aux.foldl deps_union (empty_deps, Aux.mapfilter (lookup_deps l) ids) 




(* For instance mutual recursive datatypes require this *)
fun rem_duplicates elemdata =
 let fun perm (ElemData{ids=ids1, elem=elem1, ...}, 
               ElemData{ids=ids2, elem=elem2, ...}) = 
     ((elem1 = elem2) andalso ((forall (fn x => x mem_string ids2) ids1 andalso 
                                forall (fn y => y mem_string ids1) ids2)))
 in
  distinct perm elemdata
 end

fun trancl elemdata_ =
 let val elemdata = rem_duplicates elemdata_
     fun tcls_ classes tycons consts (Deps {dep_classes, dep_tycons, dep_consts}) = 
         (case dep_classes of
           [] => (case dep_tycons of
                   [] => (case dep_consts of
                           [] => TDeps {dep_classes = classes,
                                        dep_tycons = tycons,
                                        dep_consts = consts}
                        | (c::cs) => 
                          (let val deps = (Deps {dep_classes = dep_classes,
                                                 dep_tycons = dep_tycons,
                                                 dep_consts = cs})
                           in
                            (if (c mem_string consts) 
                             then tcls_ classes tycons consts deps
                             else (case (lookup_deps elemdata (Const, c)) of
                                    NONE => tcls_ classes tycons (c::consts) deps
                                  | SOME deps' => tcls_ classes tycons (c::consts) (deps_union(deps', deps))))
                           end))
                | (tc::tcs) => 
                   (let val deps = (Deps {dep_classes = dep_classes,
                                          dep_tycons = tcs,
                                          dep_consts = dep_consts})
                    in
                     (if (tc mem_string tycons) 
                      then tcls_ classes tycons consts deps
                      else (case (lookup_deps elemdata (Tycon, tc)) of
                             NONE => tcls_ classes (tc::tycons) consts deps
                           | SOME deps' => tcls_ classes (tc::tycons) consts (deps_union(deps', deps))))
                    end))
       | (cls::cs) => 
          (let val deps = (Deps {dep_classes = cs,
                                 dep_tycons = dep_tycons,
                                 dep_consts = dep_consts})
           in
            (if (cls mem_string classes) 
             then tcls_ classes tycons consts deps
             else (case (lookup_deps elemdata (Class, cls)) of
                    NONE => tcls_ (cls::classes) tycons consts deps
                  | SOME deps' => tcls_ (cls::classes) tycons consts (deps_union(deps', deps))))
           end))
 
     fun tcls [] = []
       | tcls ((ElemData{elem, ids, data, deps, ...})::xs) =
         (let val tdeps = tcls_ [] [] [] deps
              val elemT = ElemDataT {elem = elem,
                                     ids = ids,
                                     data = data,
                                     t_deps = tdeps}
          in
           elemT::(tcls xs)
          end)
 in
   tcls elemdata
 end

(* topological sort *)

fun t_sort elemdataT =
 let fun ins_dep (a as ElemDataT {elem=elem1, ids=ids1, ...}, []) = [a]
       | ins_dep (a as ElemDataT {elem=elem1, ids=ids1, ...}, 
                 (b as ElemDataT {t_deps=TDeps {dep_classes, dep_tycons, dep_consts}, ...})::xs) =
       (case elem1 of
         Class => (case (Aux.inter_str (dep_classes, ids1)) of
                    [] => b::(ins_dep (a, xs))
                   | _ =>  a::(b::xs))
       | Tycon => (case (Aux.inter_str (dep_tycons, ids1)) of
                    [] => b::(ins_dep (a, xs))
                   | _ =>  a::(b::xs))
       | Const => (case (Aux.inter_str (dep_consts, ids1)) of
                    [] => b::(ins_dep (a, xs))
                   | _ =>  a::(b::xs)))
 in
  Aux.foldr ins_dep (elemdataT, [])
 end



(* ************************************** *)



type tdeps = Symtab.key list
val map_tdeps = Aux.mapfilter  
val empty = []

fun un_keys (a : Symtab.key list) b = (a union b) 
fun mem_key (x : Symtab.key) xs = x mem xs


fun trans_dep t =
 let val d = Symtab.dest t
     fun tc t c [] = c
       | tc t c (x::xs) =
      if (mem_key x c) then tc t c xs
      else  (case (Symtab.lookup t x) of
              SOME (v, deps) => tc t (x::c) (un_keys xs deps)
            | NONE => tc t (x::c) xs)   

     fun mk_new nt [] = nt
       | mk_new nt ((k, (v, deps))::ds) = 
         (let val cls = tc t [] deps
          in
           mk_new (Symtab.update_new (k, (v, cls)) nt) ds
           handle DUP => error "Exception in trancl"
          end)
 in
  mk_new Symtab.empty d
 end  



fun sort t xs =
 let fun ins_dep ((k, b), []) = [(k, b)]
       | ins_dep ((k, b), r as ((k', b')::xs)) =
     (case (Symtab.lookup t k') of
       SOME (v, tcls) => (if (mem_key k tcls) then ((k, b)::r)
                          else (k', b')::(ins_dep ((k, b), xs)))
     | NONE => (k', b')::(ins_dep ((k, b), xs)))

 in
  Aux.foldr ins_dep (xs, [])
 end



fun union_deps t [] = []
  | union_deps t (x::xs) = (case (Symtab.lookup t x) of
                             SOME (v, tdeps) => un_keys tdeps (union_deps t xs)
                           | NONE => (union_deps t xs))

fun extend tt t =
 let val tt' = trans_dep t
     val d = Symtab.dest tt'
     fun ext m [] = m
       | ext m ((k, (v, tdeps))::xs) = (let val tdeps' = union_deps tt tdeps
                                        in
                                          ext (Symtab.update_new (k, (v, un_keys tdeps tdeps')) m) xs
                                          handle DUP => error "Not a disjoint extension"
                                        end)
 in
  ext tt d
 end     








(* *** Theorem dependencies *** *)

fun thm_deps all_ids (nm, thm) =
 let val prf = Aux.proof_of thm
     fun find_deps MinProof r = error "incomplete proof object"
       | find_deps (PBound i) r = r
       | find_deps (Abst (s, t, p)) r = find_deps p r
       | find_deps (AbsP (s, t, p)) r = find_deps p r
       | find_deps (p % t) r = find_deps p r
       | find_deps (p %% q) r = (let val r' = find_deps p r
                                 in 
                                  find_deps q r'
                                 end) 
       | find_deps (Hyp t) r = r
       | find_deps (PAxm (s, t, tys)) r = r
       | find_deps (PThm (sl, ((nm', t, ty), f))) r =
         (let val p = (Proofterm.join_proof f) 
          in
           if ((nm = nm') orelse (nm' = "")) 
           then find_deps p r
           else (if ((nm' mem_string r) orelse not(nm' mem_string all_ids))
                 then r
                 else 
                 (case (Proofterm.strip_combt p) of
                   (PAxm _, _) => r
                 | (p', _) => find_deps p' (nm'::r)))
          end)
       | find_deps _ r = r
 in
  find_deps prf []
 end


fun thms_dep_graph thms =
 let val nms = map fst thms
     fun dep_graph [] = Symtab.empty
       | dep_graph ((nm, thm)::xs) =
        (let val deps = thm_deps nms (nm, thm)
             val _ = if false then writeln ("Adding dependencies: " ^ nm ^ " ---> " ^ (commas deps))
                     else ()
         in
          Symtab.update (nm, ("", deps)) (dep_graph xs)
         end)
 in
  dep_graph thms
 end




end